import sys
from textwrap import wrap

from shaperglot import Languages


def whatuses(options) -> None:
    """Report which languages use a particular codepoint"""

    char = options.char
    if len(char) > 1:
        try:
            if (
                char.startswith("U+")
                or char.startswith("u+")
                or char.startswith("0x")
                or char.startswith("0X")
            ):
                char = chr(int(char[2:], 16))
            else:
                char = chr(int(char, 16))
        except ValueError:
            print("Could not understand codepoint " + char)
            sys.exit(1)
    langs = Languages()
    base_langs = []
    mark_langs = []
    aux_langs = []
    for lang in langs.values():
        bases = lang.bases
        marks = lang.marks
        aux = lang.auxiliaries
        lang_key = f"{lang['name']} [{lang['id']}]".replace(" ", "\u00A0")
        if char in bases:
            base_langs.append(lang_key)
        elif char in marks:
            mark_langs.append(lang_key)
        elif char in aux:
            aux_langs.append(lang_key)

    if base_langs:
        print(f"{char} is used as a base character in the following languages:")
        for line in wrap(", ".join(sorted(base_langs)), width=75):
            print("  " + line)
        print()
    if mark_langs:
        print(f"◌{char} is used as a mark character in the following languages:")
        for line in wrap(", ".join(sorted(mark_langs)), width=75):
            print("  " + line)
        print()
    if aux_langs:
        print(f"{char} is used as an auxiliary character in the following languages:")
        for line in wrap(", ".join(sorted(aux_langs)), width=75):
            print("  " + line)



import os
import re
from collections import defaultdict
from textwrap import fill

from shaperglot import Checker, Languages


try:
    import glyphsets
except ImportError:
    glyphsets = None


def report(options) -> None:
    """Report which languages are supported by the given font"""
    checker = Checker(options.font)
    langs = Languages()
    nearly = []
    supported = []
    unsupported = []
    fixes_needed = defaultdict(set)

    lang_filter = None
    if glyphsets and options.glyphset:
        lang_filter = glyphsets.languages_per_glyphset(options.glyphset)

    if options.csv:
        print(
            "Language,Name,Supported,Bases Missing,Marks Missing,Orphaned Marks,Other"
        )

    for lang in sorted(langs.keys()):
        if options.filter and not re.search(options.filter, lang):
            continue
        if lang_filter and lang not in lang_filter:
            continue
        results = checker.check(langs[lang])

        if results.is_unknown:
            continue

        if options.csv:
            report_csv(lang, langs[lang], results)
            continue

        if results.is_success:
            supported.append(lang)
            msg = "supports"
        elif results.is_nearly_success(options.nearly):
            nearly.append(lang)
            msg = "nearly supports"
        else:
            unsupported.append(lang)
            msg = "does not fully support"

        for fixtype, things in results.unique_fixes().items():
            fixes_needed[fixtype].update(things)
        if options.group:
            continue
        print(
            f"Font {msg} language '{lang}' ({langs[lang]['name']}) ({results.score:.1f}%)"
        )

        if options.verbose and options.verbose > 1:
            for subresult in results:
                print(f" * {subresult.status_code}: {subresult.message}")

    if options.csv:
        return

    if options.group:
        show_grouped(langs, nearly, supported, unsupported)
    # Collate a useful fixing guide
    short_summary(supported, nearly, unsupported)
    if options.verbose:
        long_summary(fixes_needed, unsupported)


def show_grouped(langs, nearly, supported, unsupported):
    if supported:
        print("Supported languages")
        print("===================\n")
    for lang in supported:
        print(f"Font supports language '{lang}' ({langs[lang]['name']})")

    if nearly:
        print("\nNearly supported languages")
        print("===================\n")
    for lang in nearly:
        print(f"Font nearly supports language '{lang}' ({langs[lang]['name']})")

    if unsupported:
        print("\nUnsupported languages")
        print("====================\n")
    for lang in unsupported:
        print(f"Font does not fully support language '{lang}' ({langs[lang]['name']})")


def short_summary(supported, nearly, unsupported) -> None:
    print("\n== Summary ==\n")
    print(f"* {len(supported)+len(nearly)+len(unsupported)} languages checked")
    if supported:
        print(f"* {len(supported)} languages supported")
    if nearly:
        print(f"* {len(nearly)} languages nearly supported")


def long_summary(fixes_needed, unsupported) -> None:
    if unsupported:
        print(
            fill(
                "* Unsupported languages: " + ", ".join(unsupported),
                subsequent_indent=" " * 25,
                width=os.get_terminal_size()[0] - 2,
            )
        )
        print("\nTo add support:")
    for category, fixes in fixes_needed.items():
        plural = "s" if len(fixes) > 1 else ""
        print(f" * {category.replace('_', ' ').capitalize()}{plural}: ")
        for fix in sorted(fixes):
            print("    - " + fix)


def report_csv(langcode, lang, results) -> None:
    print(f"{langcode},\"{lang['name']}\",{results.is_success},", end="")
    missing_bases = set()
    missing_marks = set()
    missing_anchors = set()
    other_errors = set()
    for result in results:
        for problem in result.problems:
            if problem.code == "bases-missing":
                missing_bases |= set(problem.context["glyphs"])
            elif problem.code == "marks-missing":
                missing_marks |= set(problem.context["glyphs"])
            elif problem.code == "orphaned-mark":
                missing_anchors.add(
                    problem.context["base"] + "/" + problem.context["mark"]
                )
            else:
                other_errors.add(problem.code)
    print(" ".join(sorted(missing_bases)), end=",")
    print(" ".join(sorted(missing_marks)), end=",")
    print(" ".join(sorted(missing_anchors)), end=",")
    print(" ".join(sorted(other_errors)), end="\n")



from collections import defaultdict
from typing import Optional

from shaperglot import Checker, Languages, Reporter

try:
    import glyphsets
except ImportError:
    glyphsets = None


def find_lang(lang: str, langs: Languages) -> Optional[str]:
    # Find the language in the languages list; could be by ID, by name, etc.
    if lang in langs:
        return lang
    for lang_id in langs.keys():
        lang_info = langs[lang_id]
        if (
            lang_info['name'].lower() == lang.lower()
            or lang_id.lower() == lang.lower()
            or lang_info["language"].lower() == lang.lower()
            or lang_info.get("autonym", "").lower() == lang.lower()
        ):
            return lang_id
    return None


def check(options) -> None:
    """Check a particular language or languages are supported"""
    checker = Checker(options.font)
    langs = Languages()
    fixes_needed = defaultdict(set)
    lang_arg = []
    for lang in options.lang:
        if glyphsets and lang in glyphsets.defined_glyphsets():
            lang_arg.extend(glyphsets.languages_per_glyphset(lang))
        else:
            lang_arg.append(lang)

    for orig_lang in lang_arg:
        lang = find_lang(orig_lang, langs)
        if not lang:
            print(f"Language '{orig_lang}' not known")
            continue

        reporter = checker.check(langs[lang])

        if reporter.is_unknown:
            print(f"Cannot determine whether font supports language '{lang}'")
        elif reporter.is_nearly_success(options.nearly):
            print(f"Font nearly supports language '{lang}' {reporter.score:.1f}%")
            for fixtype, things in reporter.unique_fixes().items():
                fixes_needed[fixtype].update(things)
        elif reporter.is_success:
            print(f"Font supports language '{lang}'")
        else:
            print(
                f"Font does not fully support language '{lang}' {reporter.score:.1f}%"
            )

        if options.verbose and options.verbose > 1:
            for result in reporter:
                print(f" * {result.message} {result.status_code}")
        elif options.verbose or not reporter.is_success:
            for result in reporter:
                if not result.is_success:
                    print(f" * {result}")

    if fixes_needed:
        show_how_to_fix(fixes_needed)


def show_how_to_fix(reporter: Reporter):
    print("\nTo add full support to nearly-supported languages:")
    for category, fixes in reporter.items():
        plural = "s" if len(fixes) > 1 else ""
        print(f" * {category.replace('_', ' ').capitalize()}{plural}: ", end="")
        print("; ".join(sorted(fixes)))



import argparse
import sys

from shaperglot.cli.check import check
from shaperglot.cli.describe import describe
from shaperglot.cli.report import report
from shaperglot.cli.whatuses import whatuses

try:
    import glyphsets
except ImportError:
    glyphsets = None


def main(args=None) -> None:
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description="Check a font file's language coverage"
    )
    subparsers = parser.add_subparsers(help='sub-commands')

    parser_describe = subparsers.add_parser('describe', help=describe.__doc__)
    parser_describe.add_argument('--verbose', '-v', action='count')
    parser_describe.add_argument(
        'lang', metavar='LANG', help='an ISO639-3 language code'
    )
    parser_describe.set_defaults(func=describe)

    parser_check = subparsers.add_parser('check', help=check.__doc__)
    parser_check.add_argument('--verbose', '-v', action='count')
    parser_check.add_argument(
        '--nearly',
        type=int,
        help="Number of fixes left to be considered nearly supported",
        default=5,
    )
    parser_check.add_argument('font', metavar='FONT', help='the font file')
    parser_check.add_argument(
        'lang',
        metavar='LANG',
        help='one or more ISO639-3 language codes'
        + (" or glyphsets" if glyphsets else ""),
        nargs="+",
    )
    parser_check.set_defaults(func=check)

    parser_report = subparsers.add_parser('report', help=report.__doc__)
    parser_report.add_argument('font', metavar='FONT', help='the font file')
    parser_report.add_argument('--verbose', '-v', action='count')
    parser_report.add_argument(
        '--nearly',
        type=int,
        help="Number of fixes left to be considered nearly supported",
        default=5,
    )
    parser_report.add_argument('--csv', action='store_true', help="Output as CSV")
    parser_report.add_argument(
        '--group', action='store_true', help="Group by success/failure"
    )
    parser_report.add_argument(
        '--filter', type=str, help="Regular expression to filter languages"
    )
    if glyphsets:
        parser_report.add_argument(
            '--glyphset',
            help="Glyph set to use for checking",
            choices=glyphsets.defined_glyphsets(),
        )
    parser_report.set_defaults(func=report)

    parser_whatuses = subparsers.add_parser('whatuses', help=whatuses.__doc__)
    parser_whatuses.add_argument('char', metavar='CHAR', help='Character or code point')
    parser_whatuses.set_defaults(func=whatuses)

    options = parser.parse_args(args)
    if not hasattr(options, "func"):
        parser.print_help()
        sys.exit(1)
    options.func(options)


if __name__ == '__main__':  # pragma: no cover
    main()
